#doc {seq seq_histo} cmd {
#seq_histo options
#} descr {
# returns a frequency table of oligos of given size (default = 2) in the sequence
#} example {
#	% seq_histo -windowsize 3 AGUCG
#	
#}

proc seq_complement {args} {
	set seq [list_pop args]
	while {[llength $args]} {
		set opt [list_shift args]
		switch -- $opt {
			-dna {
				set type dna
			}
			-rna {
				set type rna
			}
			default {
				error "Unknown option \"$opt\""
			}
		}
	}
	if ![info exists type] {
		if [regexp {[tT]} $seq] {
			set type dna
		} elseif [regexp {[Uu]} $seq] {
			set type rna
		} else {
			set type dna
		}
	}
	switch $type {
		rna {
			set seq [string_change $seq {t u T U}]
			return [string_change [string_reverse $seq] {
				a u  u a  g c  c g
				A U  U A  G C  C G
				r y  y r  m k  k h  s b  h d  b n  v b  d h
				R Y  Y R  M K  K H  S B  H D  B N  V B  D H
			}]
		}
		dna {
			set seq [string_change $seq {u t U T}]
			return [string_change [string_reverse $seq] {
				a t  t a  g c  c g
				A T  T A  G C  C G
				r y  y r  m k  k m  s s  h d  b v  v b  d h
				R Y  Y R  M K  K M  S S  H D  B V  V B  D H
			}]
		}
	}
}

foreach base {A a U u G g C c} {
	set ::bio::sec_complementary_set($base) 1
}

foreach basepair {AU GU GC UA UG CG au gu gc ua ug cg} {
	set ::bio::sec_complementary($basepair) 1
}

proc seq_complementary {base1 base2 {unknown 1}} {
	if ![info exists ::bio::sec_complementary_set($base1)] {return $unknown}
	if ![info exists ::bio::sec_complementary_set($base2)] {return $unknown}
	if [info exists ::bio::sec_complementary($base1$base2)] {return 1} else {return 0}
}

proc seq_translate {seq {code {}}} {
	upvar #0 [bio_geneticcodetable $code] trans
	set len [string length $seq]
#	if [expr {$len%3}] {error "error translating sequence: length ($len) is not a multiple of 3"}
	set seq [string tolower $seq]
	set result ""
	for {set i 0} {$i < $len} {incr i 3} {
		set codon [string range $seq $i [expr {$i+2}]]
		append result [get trans($codon) X]
	}
	return $result
}

set ::bio::map_123 {
A Ala G Gly K Lys L Leu V Val T Thr S Ser D Asp E Glu F Phe
N Asn P Pro I Ile H His R Arg Q Gln Y Tyr C Cys M Met W Trp * ***
}
set ::bio::map_321 [string toupper {
Ala A Gly G Lys K Leu L Val V Thr T Ser S Asp D Glu E Phe F
Asn N Pro P Ile I His H Arg R Gln Q Tyr Y Cys C Met M Trp W *** *
}]

proc seq_123 {seq} {
	string map $::bio::map_123 [string toupper $seq]
}

proc seq_l123 {seq} {
	list_change [split [string toupper $seq] {}] $::bio::map_123
}

proc seq_321 {seq} {
	regsub -all " " $seq "" seq
	string map $::bio::map_321 [string toupper $seq]
}

#doc {seq seq_random_comp} cmd {
#seq_random_comp table ?size?
#} descr {
# returns a random sequence with a given composition of the given size
# If size is not given the numbers in table are used to calculate the size
# table gives the composition wanted
# <ul>
# <li>dna, rna or protein: use default frequencies for this type of sequence
# <li>as a list of characters alternating with the number or percentage wanted
# </ul>
#} example {
#	% seq_random_comp protein 100
#	
#}

proc seq_random_comp {table {size 0}} {
	if {$table eq "dna"} {
		set table {A 0.25 T 0.25 G 0.25 C 0.25}
	} elseif {$table eq "rna"} {
		set table {A 0.25 U 0.25 G 0.25 C 0.25}
	} elseif {$table eq "protein"} {
		set table [split [string trimright [file_read $::bio::dir/conf/stats/aa_freq.txt] \n] \n\t]
	}
	set chars [list_unmerge $table 1 temp]
	set total [lmath_sum $temp]
	unset -nocomplain a
	if {($size == 0) || ($total == $size)} {
		array set a $table
		set size [expr {round($total)}]
	} else {
		set tot 0
		foreach {char p} $table {
			set num [expr {round($size*double($p)/$total)}]
			set a($char) $num
			incr tot $num
		}
		if {$tot < $size} {
			set todo [expr {$size - $tot}]
			set extra 1
		} elseif {$tot > $size} {
			set todo [expr {$tot - $size}]
			set extra -1
		} else {
			set todo 0
		}
		while {$todo} {
			set rand [expr {rand()*$total}]
			set cur 0
			foreach {char p} $table {
				set cur [expr {$cur+$p}]
				if {$rand < $cur} break
			}
			incr a($char) $extra
			incr todo -1
		}
	}
	set seq ""
	set count $size
	while {$count} {
		set rand [expr {rand()*$count}]
		set cur 0
		foreach char $chars {
			incr cur $a($char)
			if {$rand < $cur} break
		}
		append seq $char
		incr a($char) -1
		if {$a($char) == 0} {
			set chars [list_remove $chars $char]
		}
		incr count -1
	}
	return $seq
}

#doc {seq seq_random_prob} cmd {
#seq_random_prob table ?size?
#} descr {
# returns a random sequence of the given size where probabilities of the characters are given
# table gives the probabilities wanted
# <ul>
# <li>dna, rna or protein: use default frequencies for this type of sequence
# <li>as a list of characters alternating with the number or percentage wanted
# </ul>
#} example {
#	% seq_random_prob protein 100
#	
#}

proc seq_random_prob {table size} {
	if {$table eq "dna"} {
		set table {A 0.25 T 0.25 G 0.25 C 0.25}
	} elseif {$table eq "rna"} {
		set table {A 0.25 U 0.25 G 0.25 C 0.25}
	} elseif {$table eq "protein"} {
		set table [split [string trimright [file_read $::bio::dir/conf/stats/aa_freq.txt] \n] \n\t]
	}
	array set a $table
	set chars [list_unmerge $table 1 temp]
	set total [lmath_sum $temp]
	set seq ""
	set count $size
	while {$count} {
		set rand [expr {rand()*$total}]
		set cur 0
		foreach char $chars {
			set cur [expr {$cur + $a($char)}]
			if {$rand < $cur} break
		}
		append seq $char
		incr count -1
	}
	return $seq
}
